@@ -6,6 +6,7 @@ import os
 import re
 import threading
 import weakref
+from urllib.parse import urlparse
 
 import httpx
 from authlib.integrations.httpx_client import AsyncOAuth2Client
@@ -63,6 +64,78 @@ def parse_range_header(range_header):
         raise ValueError("Invalid Range header format")
 
 
+def parse_msgraph_url(url_path):  # noqa: C901
+    """Parse a msgraph URL to extract site_name, drive_name, and path.
+
+    Supports formats:
+    - msgd://site_name/drive_name/path/to/file
+    - sharepoint://site_name/drive_name/path/to/file
+    - onedrive://drive_name/path/to/file
+    - msgd://site_name/drive_name
+    - msgd://site_name
+
+    Args:
+        url_path: The URL or path to parse
+
+    Returns:
+        tuple: (site_name, drive_name, path) where path defaults to "/"
+    """
+    if not url_path:
+        return None, None, "/"
+
+    # Handle URL format
+    if "://" in url_path:
+        parsed = urlparse(url_path)
+        protocol = parsed.scheme.lower()
+        path_parts = parsed.path.strip("/").split("/") if parsed.path.strip("/") else []
+
+        if protocol in ["msgd", "sharepoint"]:
+            # msgd:// and sharepoint:// format: protocol://site_name/drive_name/path
+            site_name = parsed.netloc if parsed.netloc else None
+            if not path_parts:
+                return site_name, None, "/"
+            elif len(path_parts) == 1:
+                return site_name, path_parts[0], "/"
+            else:
+                return site_name, path_parts[0], "/" + "/".join(path_parts[1:])
+
+        elif protocol == "onedrive":
+            # onedrive:// format: onedrive://drive_name/path (no site, personal OneDrive)
+            # The netloc becomes the drive name for OneDrive
+            drive_name = parsed.netloc if parsed.netloc else None
+            if not drive_name and path_parts:
+                # If no netloc, first path part is drive name
+                drive_name = path_parts[0]
+                file_path = (
+                    "/" + "/".join(path_parts[1:]) if len(path_parts) > 1 else "/"
+                )
+            else:
+                # netloc is drive name, path is file path
+                file_path = "/" + "/".join(path_parts) if path_parts else "/"
+            return None, drive_name, file_path
+
+        else:
+            # Unknown protocol, treat like msgd://
+            site_name = parsed.netloc if parsed.netloc else None
+            if not path_parts:
+                return site_name, None, "/"
+            elif len(path_parts) == 1:
+                return site_name, path_parts[0], "/"
+            else:
+                return site_name, path_parts[0], "/" + "/".join(path_parts[1:])
+    else:
+        # Handle path-only format
+        path_parts = url_path.strip("/").split("/") if url_path.strip("/") else []
+        site_name = None
+
+        if not path_parts:
+            return site_name, None, "/"
+        elif len(path_parts) == 1:
+            return site_name, path_parts[0], "/"
+        else:
+            return site_name, path_parts[0], "/" + "/".join(path_parts[1:])
+
+
 def wrap_http_not_found_exceptions(func):
     """Wrap a function that calls an HTTP request to handle 404 errors."""
 
@@ -1218,23 +1291,83 @@ class AbstractMSGraphFS(AsyncFileSystem):
 
 
 class MSGDriveFS(AbstractMSGraphFS):
-    """A filesystem that represents a SharePoint site drive as a filesystem.
+    """A unified filesystem for SharePoint sites and drives.
+
+    This class automatically handles both single-site/drive operations and multi-site operations
+    based on the parameters provided during initialization:
+
+    - Single-site mode: When site_name + drive_name or drive_id are provided
+    - Multi-site mode: When neither site_name + drive_name nor drive_id are provided
+
+    In multi-site mode, the filesystem can handle URL-based paths that specify
+    the site and drive dynamically (e.g., "msgd://SiteA/DriveB/file.txt").
+
+    Examples:
+    ---------
+    Single-site mode (traditional):
+        fs = MSGDriveFS(
+            client_id="your-client-id",
+            tenant_id="your-tenant-id",
+            client_secret="your-secret",
+            site_name="TestSite",
+            drive_name="Documents"
+        )
+        files = fs.ls("/folder/file.txt")  # operates on TestSite/Documents
+
+    Single-site mode with URL initialization:
+        fs = MSGDriveFS(
+            client_id="your-client-id",
+            tenant_id="your-tenant-id",
+            client_secret="your-secret",
+            url_path="msgd://TestSite/Documents"
+        )
+        files = fs.ls("/folder/file.txt")  # operates on TestSite/Documents
+
+    Multi-site mode:
+        fs = MSGDriveFS(
+            client_id="your-client-id",
+            tenant_id="your-tenant-id",
+            client_secret="your-secret"
+        )
+        files = fs.ls("msgd://TestSite/Documents/folder/file.txt")  # dynamic routing
+
+    Using with fsspec (recommended):
+        import fsspec
+
+        # Single-site via fsspec
+        fs = fsspec.filesystem(
+            "msgd",
+            client_id="...",
+            tenant_id="...",
+            client_secret="...",
+            site_name="TestSite",
+            drive_name="Documents"
+        )
+        files = fs.ls("/folder/")
+
+        # Multi-site via fsspec
+        fs = fsspec.filesystem("msgd", client_id="...", tenant_id="...", client_secret="...")
+        files = fs.ls("msgd://TestSite/Documents/folder/")
 
     Parameters:
     -----------
     drive_id : str, optional
-        The ID of the SharePoint drive. If not provided, will attempt to discover from site.
+        The ID of the SharePoint drive. If provided, enables single-site mode.
     client_id : str, optional
-        OAuth2 client ID. Can also be set via MSGRAPHFS_CLIENT_ID environment variable.
+        OAuth2 client ID. Can also be set via MSGRAPHFS_CLIENT_ID or AZURE_CLIENT_ID environment variables.
     tenant_id : str, optional
-        OAuth2 tenant ID. Can also be set via MSGRAPHFS_TENANT_ID environment variable.
+        OAuth2 tenant ID. Can also be set via MSGRAPHFS_TENANT_ID or AZURE_TENANT_ID environment variables.
     client_secret : str, optional
-        OAuth2 client secret. Can also be set via MSGRAPHFS_CLIENT_SECRET environment variable.
+        OAuth2 client secret. Can also be set via MSGRAPHFS_CLIENT_SECRET or AZURE_CLIENT_SECRET environment variables.
     site_name : str, optional
-        The name of the SharePoint site (optional, used for site discovery and drive resolution).
+        The name of the SharePoint site. If provided with drive_name, enables single-site mode.
     drive_name : str, optional
         The name of the SharePoint drive/library (e.g., "Documents", "CustomLibrary").
-        If not provided, uses the default drive for the site.
+        If provided with site_name, enables single-site mode.
+    url_path : str, optional
+        URL-style path specification (e.g., "msgd://TestSite/Documents").
+        If provided, extracts site_name and drive_name from the URL.
+        URL parameters override direct site_name/drive_name parameters.
     oauth2_client_params : dict, optional
         Parameters for the OAuth2 client. If not provided, will be built from client_id, tenant_id, client_secret.
     use_recycle_bin : bool, optional
@@ -1243,7 +1376,7 @@ class MSGDriveFS(AbstractMSGraphFS):
         Additional arguments passed to the parent class.
     """
 
-    protocol = ["msgd"]
+    protocol = ("msgd", "sharepoint", "onedrive")
 
     # Default OAuth2 scopes for Microsoft Graph API (client credentials flow)
     DEFAULT_SCOPES = ["https://graph.microsoft.com/.default"]
@@ -1259,12 +1392,50 @@ class MSGDriveFS(AbstractMSGraphFS):
         oauth2_client_params: dict | None = None,
         asynchronous: bool = False,
         loop=None,
+        url_path: str | None = None,
         **kwargs,
     ):
         # Get OAuth2 credentials from parameters or environment variables
-        self.client_id = client_id or os.getenv("MSGRAPHFS_CLIENT_ID")
-        self.tenant_id = tenant_id or os.getenv("MSGRAPHFS_TENANT_ID")
-        self.client_secret = client_secret or os.getenv("MSGRAPHFS_CLIENT_SECRET")
+        # Check MSGRAPHFS_* variables first, then fall back to standard AZURE_* variables
+        self.client_id = (
+            client_id
+            or os.getenv("MSGRAPHFS_CLIENT_ID")
+            or os.getenv("AZURE_CLIENT_ID")
+        )
+        self.tenant_id = (
+            tenant_id
+            or os.getenv("MSGRAPHFS_TENANT_ID")
+            or os.getenv("AZURE_TENANT_ID")
+        )
+        self.client_secret = (
+            client_secret
+            or os.getenv("MSGRAPHFS_CLIENT_SECRET")
+            or os.getenv("AZURE_CLIENT_SECRET")
+        )
+
+        # Parse URL path if provided to extract site_name and drive_name
+        if url_path:
+            parsed_site, parsed_drive, _ = parse_msgraph_url(url_path)
+            # URL parameters override direct parameters
+            site_name = parsed_site or site_name
+            drive_name = parsed_drive or drive_name
+
+        # Determine operation mode (single-site if site and drive provided, OR drive_id provided)
+        self._multi_site_mode = not ((site_name and drive_name) or drive_id)
+
+        # Set site_name and drive_name attributes for all modes
+        self.site_name = site_name
+        self.drive_name = drive_name
+
+        if self._multi_site_mode:
+            # Multi-site mode: cache for drive filesystem instances
+            self._drive_cache = {}
+            # Store credentials for creating drive instances
+            self._stored_client_id = self.client_id
+            self._stored_tenant_id = self.tenant_id
+            self._stored_client_secret = self.client_secret
+            self._stored_oauth2_client_params = oauth2_client_params
+            self._stored_kwargs = kwargs.copy()
 
         # Build oauth2_client_params if not provided
         if oauth2_client_params is None:
@@ -1272,8 +1443,9 @@ class MSGDriveFS(AbstractMSGraphFS):
                 raise ValueError(
                     "Either oauth2_client_params must be provided, or all of "
                     "client_id, tenant_id, and client_secret must be provided "
-                    "(either as parameters or environment variables MSGRAPHFS_CLIENT_ID, "
-                    "MSGRAPHFS_TENANT_ID, MSGRAPHFS_CLIENT_SECRET)"
+                    "(either as parameters or environment variables MSGRAPHFS_CLIENT_ID/"
+                    "AZURE_CLIENT_ID, MSGRAPHFS_TENANT_ID/AZURE_TENANT_ID, "
+                    "MSGRAPHFS_CLIENT_SECRET/AZURE_CLIENT_SECRET)"
                 )
 
             # Build OAuth2 client parameters with proper configuration
@@ -1309,6 +1481,16 @@ class MSGDriveFS(AbstractMSGraphFS):
         else:
             self.drive_url = None
 
+    def _parse_path_for_url_routing(self, path: str):
+        """Parse a path to extract site_name, drive_name, and file path for URL
+        routing."""
+        site_name, drive_name, file_path = parse_msgraph_url(path)
+        if not site_name:
+            raise ValueError(f"Path must include site name: {path}")
+        if not drive_name:
+            raise ValueError(f"Path must include drive name: {path}")
+        return site_name, drive_name, file_path
+
     def _extract_tenant_from_token_endpoint(self, token_endpoint: str) -> str | None:
         """Extract tenant_id from token endpoint URL."""
         import re
@@ -1316,6 +1498,152 @@ class MSGDriveFS(AbstractMSGraphFS):
         match = re.search(r"/([a-f0-9-]+)/oauth2", token_endpoint)
         return match.group(1) if match else None
 
+    def _parse_path_for_missing_components(self, path: str):
+        """Parse a path to extract missing site_name, drive_name, and return the file
+        path.
+
+        Logic:
+        - If both site_name and drive_name are set: path is the file path
+        - If only site_name is set: path = drive_name/file_path
+        - If neither is set: path = site_name/drive_name/file_path
+        """
+        # If we have both site and drive, no parsing needed
+        if self.site_name and self.drive_name:
+            return self.site_name, self.drive_name, path
+
+        # Parse the path to extract missing components
+        if "://" in path:
+            # Handle URL format
+            parsed_site, parsed_drive, file_path = parse_msgraph_url(path)
+            # For OneDrive URLs, use a default site name if none specified
+            if parsed_site is None and "onedrive://" in path.lower():
+                site_name = self.site_name or "me"  # Use "me" as default OneDrive site
+            else:
+                site_name = self.site_name or parsed_site
+            drive_name = self.drive_name or parsed_drive
+        else:
+            # Handle plain path format
+            if not self.site_name and not self.drive_name:
+                # Need both site and drive from path: site/drive/file_path
+                path_parts = path.strip("/").split("/", 2)
+                if len(path_parts) < 2:
+                    raise ValueError(
+                        f"Path must include site and drive when none specified: {path}"
+                    )
+                site_name = path_parts[0]
+                drive_name = path_parts[1]
+                file_path = "/" + path_parts[2] if len(path_parts) > 2 else "/"
+            elif self.site_name and not self.drive_name:
+                # Need drive from path: drive/file_path
+                path_parts = path.strip("/").split("/", 1)
+                if len(path_parts) < 1:
+                    raise ValueError(
+                        f"Path must include drive name when not specified: {path}"
+                    )
+                site_name = self.site_name
+                drive_name = path_parts[0]
+                file_path = "/" + path_parts[1] if len(path_parts) > 1 else "/"
+            else:
+                # This shouldn't happen but handle gracefully
+                site_name = self.site_name
+                drive_name = self.drive_name
+                file_path = path
+
+        if not site_name or not drive_name:
+            raise ValueError(f"Unable to determine site and drive from path: {path}")
+
+        return site_name, drive_name, file_path
+
+    def _get_drive_fs(self, site_name: str, drive_name: str) -> "MSGDriveFS":
+        """Get or create a MSGDriveFS instance for the specified site and drive."""
+        # If this instance already has the right site/drive, return self
+        if self.site_name == site_name and self.drive_name == drive_name:
+            return self
+
+        # If we have a drive cache, use it (always available in multi-site mode)
+        if hasattr(self, "_drive_cache") and self._drive_cache is not None:
+            cache_key = (site_name, drive_name)
+            if cache_key not in self._drive_cache:
+                self._drive_cache[cache_key] = MSGDriveFS(
+                    client_id=self.client_id,
+                    tenant_id=self.tenant_id,
+                    client_secret=self.client_secret,
+                    site_name=site_name,
+                    drive_name=drive_name,
+                    asynchronous=self.asynchronous,
+                    loop=self.loop,
+                )
+            return self._drive_cache[cache_key]
+
+        # No caching needed, create a new instance
+        return MSGDriveFS(
+            client_id=self.client_id,
+            tenant_id=self.tenant_id,
+            client_secret=self.client_secret,
+            site_name=site_name,
+            drive_name=drive_name,
+            asynchronous=self.asynchronous,
+            loop=self.loop,
+        )
+
+    # Delegation methods for multi-site operations (used when _multi_site_mode is True)
+    async def _ls_multi_site(self, path: str, detail: bool = True, **kwargs):
+        """List files in multi-site mode by delegating to appropriate drive
+        filesystem."""
+        site_name, drive_name, file_path = self._parse_path_for_url_routing(path)
+        drive_fs = self._get_drive_fs(site_name, drive_name)
+        return await drive_fs._ls(file_path, detail=detail, **kwargs)
+
+    async def _info_multi_site(self, path: str, **kwargs):
+        """Get file info in multi-site mode by delegating to appropriate drive
+        filesystem."""
+        site_name, drive_name, file_path = self._parse_path_for_url_routing(path)
+        drive_fs = self._get_drive_fs(site_name, drive_name)
+        return await drive_fs._info(file_path, **kwargs)
+
+    async def _cat_file_multi_site(
+        self, path: str, start: int | None = None, end: int | None = None, **kwargs
+    ):
+        """Read file content in multi-site mode by delegating to appropriate drive
+        filesystem."""
+        site_name, drive_name, file_path = self._parse_path_for_url_routing(path)
+        drive_fs = self._get_drive_fs(site_name, drive_name)
+        return await drive_fs._cat_file(file_path, start=start, end=end, **kwargs)
+
+    def _open_multi_site(self, path: str, mode: str = "rb", **kwargs):
+        """Open file in multi-site mode by delegating to appropriate drive
+        filesystem."""
+        site_name, drive_name, file_path = self._parse_path_for_url_routing(path)
+        drive_fs = self._get_drive_fs(site_name, drive_name)
+        return drive_fs._open(file_path, mode=mode, **kwargs)
+
+    # Override existing methods to delegate to multi-site variants when needed
+    async def _ls(self, path: str, detail: bool = True, **kwargs):
+        """List files, delegating to multi-site logic if needed."""
+        if self._multi_site_mode:
+            return await self._ls_multi_site(path, detail=detail, **kwargs)
+        return await super()._ls(path, detail=detail, **kwargs)
+
+    async def _info(self, path: str, **kwargs):
+        """Get file info, delegating to multi-site logic if needed."""
+        if self._multi_site_mode:
+            return await self._info_multi_site(path, **kwargs)
+        return await super()._info(path, **kwargs)
+
+    async def _cat_file(
+        self, path: str, start: int | None = None, end: int | None = None, **kwargs
+    ):
+        """Read file content, delegating to multi-site logic if needed."""
+        if self._multi_site_mode:
+            return await self._cat_file_multi_site(path, start=start, end=end, **kwargs)
+        return await super()._cat_file(path, start=start, end=end, **kwargs)
+
+    def _open(self, path: str, mode: str = "rb", **kwargs):
+        """Open file, delegating to multi-site logic if needed."""
+        if self._multi_site_mode:
+            return self._open_multi_site(path, mode=mode, **kwargs)
+        return super()._open(path, mode=mode, **kwargs)
+
     async def _ensure_drive_id(self) -> str:
         """Ensure drive_id is available, discovering it if necessary."""
         if self.drive_id:
@@ -1438,6 +1766,87 @@ class MSGDriveFS(AbstractMSGraphFS):
     ensure_drive_id = sync_wrapper(_ensure_drive_id)
     get_drive_id_by_name = sync_wrapper(_get_drive_id_by_name)
 
+    # Override filesystem operations to support path-based site/drive resolution
+    async def _ls(self, path: str, detail: bool = True, **kwargs):
+        """List files, supporting path-based site/drive resolution."""
+        site_name, drive_name, file_path = self._parse_path_for_missing_components(path)
+        drive_fs = self._get_drive_fs(site_name, drive_name)
+
+        # If we got back ourselves, use normal behavior
+        if drive_fs is self:
+            return await super()._ls(file_path, detail=detail, **kwargs)
+        else:
+            # Delegate to the appropriate drive filesystem
+            return await drive_fs._ls(file_path, detail=detail, **kwargs)
+
+    async def _info(self, path: str, **kwargs):
+        """Get file info, supporting path-based site/drive resolution."""
+        site_name, drive_name, file_path = self._parse_path_for_missing_components(path)
+        drive_fs = self._get_drive_fs(site_name, drive_name)
+
+        if drive_fs is self:
+            return await super()._info(file_path, **kwargs)
+        else:
+            return await drive_fs._info(file_path, **kwargs)
+
+    async def _cat_file(
+        self, path: str, start: int | None = None, end: int | None = None, **kwargs
+    ):
+        """Read file content, supporting path-based site/drive resolution."""
+        site_name, drive_name, file_path = self._parse_path_for_missing_components(path)
+        drive_fs = self._get_drive_fs(site_name, drive_name)
+
+        if drive_fs is self:
+            return await super()._cat_file(file_path, start=start, end=end, **kwargs)
+        else:
+            return await drive_fs._cat_file(file_path, start=start, end=end, **kwargs)
+
+    def _open(self, path: str, mode: str = "rb", **kwargs):
+        """Open file, supporting path-based site/drive resolution."""
+        site_name, drive_name, file_path = self._parse_path_for_missing_components(path)
+        drive_fs = self._get_drive_fs(site_name, drive_name)
+
+        if drive_fs is self:
+            return super()._open(file_path, mode=mode, **kwargs)
+        else:
+            return drive_fs._open(file_path, mode=mode, **kwargs)
+
+    async def _exists(self, path: str, **kwargs):
+        """Check if file exists, supporting path-based site/drive resolution."""
+        site_name, drive_name, file_path = self._parse_path_for_missing_components(path)
+        drive_fs = self._get_drive_fs(site_name, drive_name)
+
+        if drive_fs is self:
+            return await super()._exists(file_path, **kwargs)
+        else:
+            return await drive_fs._exists(file_path, **kwargs)
+
+    async def _mkdir(
+        self, path: str, create_parents: bool = True, exist_ok: bool = False, **kwargs
+    ):
+        """Create directory, supporting path-based site/drive resolution."""
+        site_name, drive_name, file_path = self._parse_path_for_missing_components(path)
+        drive_fs = self._get_drive_fs(site_name, drive_name)
+
+        if drive_fs is self:
+            return await super()._mkdir(
+                file_path, create_parents=create_parents, exist_ok=exist_ok, **kwargs
+            )
+        else:
+            return await drive_fs._mkdir(
+                file_path, create_parents=create_parents, exist_ok=exist_ok, **kwargs
+            )
+
+    async def _rm(self, path: str, recursive: bool = False, **kwargs):
+        """Remove file/directory, supporting path-based site/drive resolution."""
+        site_name, drive_name, file_path = self._parse_path_for_missing_components(path)
+        drive_fs = self._get_drive_fs(site_name, drive_name)
+
+        if drive_fs is self:
+            return await super()._rm(file_path, recursive=recursive, **kwargs)
+        else:
+            return await drive_fs._rm(file_path, recursive=recursive, **kwargs)
+
 
 class AsyncStreamedFileMixin:
     """Mixin for streamed file-like objects using async iterators."""
